# Copyright (2015-2017) Hewlett Packard Enterprise Development LP
# Copyright (2015-2017) Universidade Federal de Campina Grande
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import re

from oslo_log import log as logging
from oslo_serialization import jsonutils
from oslo_utils import importutils
from six.moves.urllib import parse

from ironic.common import exception
from ironic.common.i18n import _
from ironic.conf import CONF
from ironic.drivers import utils

LOG = logging.getLogger(__name__)

hponeview_client = importutils.try_import('hpOneView.oneview_client')
redfish = importutils.try_import('redfish')
client_exception = importutils.try_import('hpOneView.exceptions')

REQUIRED_ON_DRIVER_INFO = {
    'server_hardware_uri': _("Server Hardware URI. Required in driver_info."),
}

REQUIRED_ON_PROPERTIES = {
    'server_hardware_type_uri': _(
        "Server Hardware Type URI. Required in properties/capabilities."
    ),
    'server_profile_template_uri': _(
        "Server Profile Template URI to clone from. "
        "Required in properties/capabilities."
    ),
}

OPTIONAL_ON_PROPERTIES = {
    'enclosure_group_uri': _(
        "Enclosure Group URI. Optional in properties/capabilities."),
}

ILOREST_BASE_PORT = "443"

COMMON_PROPERTIES = {}
COMMON_PROPERTIES.update(REQUIRED_ON_DRIVER_INFO)
COMMON_PROPERTIES.update(REQUIRED_ON_PROPERTIES)
COMMON_PROPERTIES.update(OPTIONAL_ON_PROPERTIES)

ISCSI_PXE_ONEVIEW = 'iscsi_pxe_oneview'
AGENT_PXE_ONEVIEW = 'agent_pxe_oneview'

# NOTE(xavierr): We don't want to translate NODE_IN_USE_BY_ONEVIEW and
# SERVER_HARDWARE_ALLOCATION_ERROR to avoid inconsistency in the nodes
# caused by updates on translation in upgrades of ironic.
NODE_IN_USE_BY_ONEVIEW = 'node in use by OneView'
SERVER_HARDWARE_ALLOCATION_ERROR = 'server hardware allocation error'


def prepare_manager_url(manager_url):
    # NOTE(mrtenio) python-oneviewclient uses https or http in the manager_url
    # while python-hpOneView does not. This will not be necessary when
    # python-hpOneView client is the only OneView library.
    if manager_url:
        url_match = "^(http[s]?://)?([^/]+)(/.*)?$"
        manager_url = re.search(url_match, manager_url).group(2)
    return manager_url


def get_hponeview_client():
    """Generate an instance of the hpOneView client.

    Generates an instance of the hpOneView client using the hpOneView library.

    :returns: an instance of the OneViewClient
    :raises: InvalidParameterValue if mandatory information is missing on the
             node or on invalid input.
    :raises: OneViewError if try a secure connection without CA certificate.
    """
    manager_url = prepare_manager_url(CONF.oneview.manager_url)

    insecure = CONF.oneview.allow_insecure_connections
    ssl_certificate = CONF.oneview.tls_cacert_file

    if not (insecure or ssl_certificate):
        msg = _("TLS CA certificate to connect with OneView is missing.")
        raise exception.OneViewError(error=msg)

    # NOTE(nicodemos) Ignore the CA certificate if it's an insecure connection
    if insecure and ssl_certificate:
        LOG.warning("Performing an insecure connection with OneView, the CA "
                    "certificate file: %s will be ignored.", ssl_certificate)
        ssl_certificate = None

    config = {
        "ip": manager_url,
        "credentials": {
            "userName": CONF.oneview.username,
            "password": CONF.oneview.password
        },
        "ssl_certificate": ssl_certificate
    }
    return hponeview_client.OneViewClient(config)


def get_ilorest_client(server_hardware):
    """Generate an instance of the iLORest library client.

    :param: server_hardware: a server hardware uuid or uri
    :returns: an instance of the iLORest client
    :raises: InvalidParameterValue if mandatory information is missing on the
             node or on invalid input.
    """
    oneview_client = get_hponeview_client()
    remote_console = oneview_client.server_hardware.get_remote_console_url(
        server_hardware
    )
    host_ip, ilo_token = _get_ilo_access(remote_console)
    base_url = "https://%s:%s" % (host_ip, ILOREST_BASE_PORT)
    return redfish.rest_client(base_url=base_url, sessionkey=ilo_token)


def _get_ilo_access(remote_console):
    """Get the needed information to access ilo.

    Get the host_ip and a token of an iLO remote console instance which can be
    used to perform operations on that controller.

    The Remote Console url has the following format:
    hplocons://addr=1.2.3.4&sessionkey=a79659e3b3b7c8209c901ac3509a6719

    :param: remote_console: OneView Remote Console object with a
            remoteConsoleUrl
    :returns: A tuple with the Host IP and Token to access ilo, for
              example: ('1.2.3.4', 'a79659e3b3b7c8209c901ac3509a6719')
    """
    url = remote_console.get('remoteConsoleUrl')
    url_parse = parse.urlparse(url)
    host_ip = parse.parse_qs(url_parse.netloc).get('addr')[0]
    token = parse.parse_qs(url_parse.netloc).get('sessionkey')[0]
    return host_ip, token


def verify_node_info(node):
    """Verifies if fields and namespaces of a node are valid.

    Verifies if the 'driver_info' field and the 'properties/capabilities'
    namespace exist and are not empty.

    :param: node: node object to be verified
    :raises: InvalidParameterValue if required node capabilities and/or
             driver_info are malformed or missing
    :raises: MissingParameterValue if required node capabilities and/or
             driver_info are missing
    """
    capabilities_dict = utils.capabilities_to_dict(
        node.properties.get('capabilities', '')
    )
    driver_info = node.driver_info

    _verify_node_info('properties/capabilities', capabilities_dict,
                      REQUIRED_ON_PROPERTIES)

    _verify_node_info('driver_info', driver_info,
                      REQUIRED_ON_DRIVER_INFO)


def get_oneview_info(node):
    """Gets OneView information from the node.

    :param: node: node object to get information from
    :returns: a dictionary containing:
        :server_hardware_uri: the uri of the server hardware in OneView
        :server_hardware_type_uri: the uri of the server hardware type in
            OneView
        :enclosure_group_uri: the uri of the enclosure group in OneView
        :server_profile_template_uri: the uri of the server profile template in
            OneView
    :raises OneViewInvalidNodeParameter: if node capabilities are malformed
    """

    try:
        capabilities_dict = utils.capabilities_to_dict(
            node.properties.get('capabilities', '')
        )
    except exception.InvalidParameterValue as e:
        raise exception.OneViewInvalidNodeParameter(node_uuid=node.uuid,
                                                    error=e)

    driver_info = node.driver_info

    oneview_info = {
        'server_hardware_uri':
            driver_info.get('server_hardware_uri'),
        'server_hardware_type_uri':
            capabilities_dict.get('server_hardware_type_uri'),
        'enclosure_group_uri':
            capabilities_dict.get('enclosure_group_uri'),
        'server_profile_template_uri':
            capabilities_dict.get('server_profile_template_uri'),
        'applied_server_profile_uri':
            driver_info.get('applied_server_profile_uri'),
    }

    return oneview_info


def validate_oneview_resources_compatibility(task):
    """Validate if the node configuration is consistent with OneView.

    This method calls hpOneView functions to validate if the node
    configuration is consistent with the OneView resources it represents,
    including serverHardwareUri, serverHardwareTypeUri, serverGroupUri
    serverProfileTemplateUri, enclosureGroupUri and node ports. If any
    validation fails, the driver will raise an appropriate OneViewError.

    :param: task: a TaskManager instance containing the node to act on.
    :raises: OneViewError if any validation fails.
    """
    ports = task.ports
    oneview_client = get_hponeview_client()
    oneview_info = get_oneview_info(task.node)

    _validate_node_server_profile_template(oneview_client, oneview_info)
    _validate_node_server_hardware_type(oneview_client, oneview_info)
    _validate_node_enclosure_group(oneview_client, oneview_info)
    _validate_server_profile_template_mac_type(oneview_client, oneview_info)
    _validate_node_port_mac_server_hardware(
        oneview_client, oneview_info, ports)


def _verify_node_info(node_namespace, node_info_dict, info_required):
    """Verify if info_required is present in node_namespace of the node info.

    """
    missing_keys = set(info_required) - set(node_info_dict)

    if missing_keys:
        raise exception.MissingParameterValue(
            _("Missing the keys for the following OneView data in node's "
              "%(namespace)s: %(missing_keys)s.") %
            {'namespace': node_namespace,
             'missing_keys': ', '.join(missing_keys)
             }
        )

    # False and 0 can still be considered as valid values
    missing_values_keys = [k for k in info_required
                           if node_info_dict[k] in ('', None)]
    if missing_values_keys:
        missing_keys = ["%s:%s" % (node_namespace, k)
                        for k in missing_values_keys]
        raise exception.MissingParameterValue(
            _("Missing parameter value for: '%s'") % "', '".join(missing_keys)
        )


def node_has_server_profile(func):
    """Checks if the node's Server Hardware has a Server Profile associated.

    Decorator to execute before the function execution if the Server Profile
    is applied to the Server Hardware.

    :param func: a given decorated function.
    """
    def inner(self, *args, **kwargs):
        task = args[0]
        ensure_server_profile(task)
        return func(self, *args, **kwargs)
    return inner


def ensure_server_profile(task):
    """Checks if the node's Server Hardware has a Server Profile associated.

    Function to check if the Server Profile is applied to the Server Hardware.

    :param task: a TaskManager instance containing the node to act on.
    :raises: OneViewError if failed to get server profile from OneView
    """
    oneview_client = get_hponeview_client()
    try:
        profile_uri = task.node.driver_info.get('applied_server_profile_uri')
        oneview_client.server_profiles.get(profile_uri)
    except client_exception.HPOneViewException as exc:
        LOG.error(
            "Failed to get server profile: %(profile)s from OneView appliance "
            "for node %(node)s. Error: %(message)s", {
                "profile": profile_uri,
                "node": task.node.uuid,
                "message": exc
            }
        )
        raise exception.OneViewError(error=exc)


def _get_server_hardware_mac_from_ilo(server_hardware):
    """Get the MAC of Server Hardware's iLO controller.

    :param: server_hardware: a server hardware uuid or uri
    :return: MAC of Server Hardware's iLO controller.
    :raises: InvalidParameterValue if required iLO credentials are missing.
    :raises: OneViewError if can't get mac from a server hardware via iLO or
             if fails to get JSON object with the default path.
    """
    try:
        ilo_client = get_ilorest_client(server_hardware)
        ilo_path = "/rest/v1/systems/1"
        hardware = jsonutils.loads(ilo_client.get(ilo_path).text)
        hardware_mac = hardware['HostCorrelation']['HostMACAddress'][0]
    except redfish.JsonDecodingError as exc:
        LOG.error("Failed in JSON object getting path: %s", ilo_path)
        raise exception.OneViewError(error=exc)
    except (ValueError, TypeError, IndexError) as exc:
        LOG.exception(
            "Failed to get mac from server hardware %(server_hardware)s "
            "via iLO. Error: %(message)s", {
                "server_hardware": server_hardware.get("uri"),
                "message": exc
            }
        )
        raise exception.OneViewError(error=exc)

    return hardware_mac


def _get_server_hardware_mac(server_hardware):
    """Get the MAC address of the first PXE bootable port of an Ethernet port.

    :param: server_hardware: OneView Server Hardware object.
    :return: MAC of the first Ethernet and function 'a' port of the
             Server Hardware object.
    :raises: OneViewError if there is no Ethernet port on the Server Hardware
             or if there is no portMap on the Server Hardware requested.
    """
    sh_physical_port = None

    if server_hardware.get('portMap'):
        for device in server_hardware.get(
                'portMap', {}).get('deviceSlots', ()):
            for physical_port in device.get('physicalPorts', ()):
                if physical_port.get('type') == 'Ethernet':
                    sh_physical_port = physical_port
                    break
        if sh_physical_port:
            for virtual_port in sh_physical_port.get('virtualPorts', ()):
                # NOTE(nicodemos): Ironic oneview drivers needs to use a
                # port that type is Ethernet and function identifier 'a' for
                # this FlexNIC to be able to make a deploy using PXE.
                if virtual_port.get('portFunction') == 'a':
                    return virtual_port.get('mac', ()).lower()
        raise exception.OneViewError(
            _("There is no Ethernet port on the Server Hardware: %s") %
            server_hardware.get('uri'))
    else:
        raise exception.OneViewError(
            _("The Server Hardware: %s doesn't have a list of adapters/slots, "
              "their ports and attributes. This information is available only "
              "for blade servers. Is this a rack server?") %
            server_hardware.get('uri'))


def _validate_node_server_profile_template(oneview_client, oneview_info):
    """Validate if the Server Profile Template is consistent.

    :param: oneview_client: an instance of the HPE OneView client.
    :param: oneview_info: the OneView related info in an Ironic node.
    :raises: OneViewError if the node's Server Profile Template is not
             consistent.
    """
    server_profile_template = oneview_client.server_profile_templates.get(
        oneview_info['server_profile_template_uri'])
    server_hardware = oneview_client.server_hardware.get(
        oneview_info['server_hardware_uri'])

    _validate_server_profile_template_server_hardware_type(
        server_profile_template, server_hardware)
    _validate_spt_enclosure_group(server_profile_template, server_hardware)
    _validate_server_profile_template_manage_boot(server_profile_template)


def _validate_server_profile_template_server_hardware_type(
        server_profile_template, server_hardware):
    """Validate if the Server Hardware Types are the same.

    Validate if the Server Profile Template and the Server Hardware have the
    same Server Hardware Type.

    :param: server_profile_template: OneView Server Profile Template object.
    :param: server_hardware: OneView Server Hardware object.
    :raises: OneViewError if the Server Profile Template and the Server
             Hardware does not have the same Server Hardware Type.
    """
    spt_server_hardware_type_uri = (
        server_profile_template.get('serverHardwareTypeUri')
    )
    sh_server_hardware_type_uri = server_hardware.get('serverHardwareTypeUri')

    if spt_server_hardware_type_uri != sh_server_hardware_type_uri:
        message = _(
            "Server profile template %(spt_uri)s serverHardwareTypeUri is "
            "inconsistent with server hardware %(server_hardware_uri)s "
            "serverHardwareTypeUri.") % {
                'spt_uri': server_profile_template.get('uri'),
                'server_hardware_uri': server_hardware.get('uri')}
        raise exception.OneViewError(message)


def _validate_spt_enclosure_group(server_profile_template, server_hardware):
    """Validate Server Profile Template's Enclosure Group and Server Hardware's.

    :param: server_profile_template: OneView Server Profile Template object.
    :param: server_hardware: OneView Server Hardware object.
    :raises: OneViewError if the Server Profile Template's Enclosure Group does
             not match the Server Hardware's.
    """
    spt_enclosure_group_uri = server_profile_template.get('enclosureGroupUri')
    sh_enclosure_group_uri = server_hardware.get('serverGroupUri')

    if spt_enclosure_group_uri != sh_enclosure_group_uri:
        message = _("Server profile template %(spt_uri)s enclosureGroupUri is "
                    "inconsistent with server hardware %(sh_uri)s "
                    "serverGroupUri.") % {
                        'spt_uri': server_profile_template.get('uri'),
                        'sh_uri': server_hardware.get('uri')}
        raise exception.OneViewError(message)


def _validate_server_profile_template_manage_boot(server_profile_template):
    """Validate if the Server Profile Template allows to manage the boot order.

    :param: server_profile_template: OneView Server Profile Template object.
    :raises: OneViewError if the Server Profile Template does not allows to
             manage the boot order.
    """
    manage_boot = server_profile_template.get('boot', {}).get('manageBoot')

    if not manage_boot:
        message = _("Server Profile Template: %s, does not allow to manage "
                    "boot order.") % server_profile_template.get('uri')
        raise exception.OneViewError(message)


def _validate_node_server_hardware_type(oneview_client, oneview_info):
    """Validate if the node's Server Hardware Type matches Server Hardware's.

    :param: oneview_client: the HPE OneView Client.
    :param: oneview_info: the OneView related info in an Ironic node.
    :raises: OneViewError if the node's Server Hardware Type group doesn't
             match the Server Hardware's.
    """
    node_server_hardware_type_uri = oneview_info['server_hardware_type_uri']
    server_hardware = oneview_client.server_hardware.get(
        oneview_info['server_hardware_uri'])
    server_hardware_sht_uri = server_hardware.get('serverHardwareTypeUri')

    if server_hardware_sht_uri != node_server_hardware_type_uri:
        message = _("Node server_hardware_type_uri is inconsistent "
                    "with OneView's server hardware %(server_hardware_uri)s "
                    "serverHardwareTypeUri.") % {
                        'server_hardware_uri': server_hardware.get('uri')}
        raise exception.OneViewError(message)


def _validate_node_enclosure_group(oneview_client, oneview_info):
    """Validate if the node's Enclosure Group matches the Server Hardware's.

    :param: oneview_client: an instance of the HPE OneView client.
    :param: oneview_info: the OneView related info in an Ironic node.
    :raises: OneViewError if the node's enclosure group doesn't match the
             Server Hardware's.
    """
    server_hardware = oneview_client.server_hardware.get(
        oneview_info['server_hardware_uri'])
    sh_enclosure_group_uri = server_hardware.get('serverGroupUri')
    node_enclosure_group_uri = oneview_info['enclosure_group_uri']

    if node_enclosure_group_uri and (
            sh_enclosure_group_uri != node_enclosure_group_uri):
        message = _(
            "Node enclosure_group_uri '%(node_enclosure_group_uri)s' "
            "is inconsistent with OneView's server hardware "
            "serverGroupUri '%(sh_enclosure_group_uri)s' of "
            "ServerHardware %(server_hardware)s") % {
                'node_enclosure_group_uri': node_enclosure_group_uri,
                'sh_enclosure_group_uri': sh_enclosure_group_uri,
                'server_hardware': server_hardware.get('uri')}
        raise exception.OneViewError(message)


def _validate_node_port_mac_server_hardware(oneview_client,
                                            oneview_info, ports):
    """Validate if a port matches the node's Server Hardware's MAC.

    :param: oneview_client: an instance of the HPE OneView client.
    :param: oneview_info: the OneView related info in an Ironic node.
    :param: ports: a list of Ironic node's ports.
    :raises: OneViewError if there is no port with MAC address matching one
    in OneView.

    """
    server_hardware = oneview_client.server_hardware.get(
        oneview_info['server_hardware_uri'])

    if not ports:
        return

    # NOTE(nicodemos) If hponeview client's unable to get the MAC of the Server
    # Hardware and raises an exception, the driver will try to get it from
    # the iLOrest client.
    try:
        mac = _get_server_hardware_mac(server_hardware)
    except exception.OneViewError:
        mac = _get_server_hardware_mac_from_ilo(server_hardware)

    incompatible_macs = []
    for port in ports:
        if port.address.lower() == mac.lower():
            return
        incompatible_macs.append(port.address)

    message = _("The ports of the node are not compatible with its "
                "server hardware %(server_hardware_uri)s. There are no Ironic "
                "port MAC's: %(port_macs)s, that matches with the "
                "server hardware's MAC: %(server_hardware_mac)s") % {
                    'server_hardware_uri': server_hardware.get('uri'),
                    'port_macs': ', '.join(incompatible_macs),
                    'server_hardware_mac': mac}
    raise exception.OneViewError(message)


def _validate_server_profile_template_mac_type(oneview_client, oneview_info):
    """Validate if the node's Server Profile Template's MAC type is physical.

    :param: oneview_client: an instance of the HPE OneView client.
    :param: oneview_info: the OneView related info in an Ironic node.
    :raises: OneViewError if the node's Server Profile Template's MAC type is
             not physical.
    """
    server_profile_template = oneview_client.server_profile_templates.get(
        oneview_info['server_profile_template_uri']
    )
    if server_profile_template.get('macType') != 'Physical':
        message = _("The server profile template %s is not set to use "
                    "physical MAC.") % server_profile_template.get('uri')
        raise exception.OneViewError(message)



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

"""
Hardware types for Cisco UCS Servers
"""

from ironic.drivers import ipmi
from ironic.drivers.modules.cimc import management as cimc_mgmt
from ironic.drivers.modules.cimc import power as cimc_power
from ironic.drivers.modules.ucs import management as ucs_mgmt
from ironic.drivers.modules.ucs import power as ucs_power


class CiscoUCSStandalone(ipmi.IPMIHardware):
    """Cisco UCS in standalone mode"""

    @property
    def supported_management_interfaces(self):
        """List of supported management interfaces."""
        mgmt = super(CiscoUCSStandalone, self).supported_management_interfaces
        return [cimc_mgmt.CIMCManagement] + mgmt

    @property
    def supported_power_interfaces(self):
        """List of supported power interfaces."""
        power = super(CiscoUCSStandalone, self).supported_power_interfaces
        return [cimc_power.Power] + power


class CiscoUCSManaged(CiscoUCSStandalone):
    """Cisco UCS under UCSM management"""

    @property
    def supported_management_interfaces(self):
        """List of supported management interfaces."""
        mgmt = super(CiscoUCSManaged, self).supported_management_interfaces
        return [ucs_mgmt.UcsManagement] + mgmt

    @property
    def supported_power_interfaces(self):
        """List of supported power interfaces."""
        power = super(CiscoUCSManaged, self).supported_power_interfaces
        return [ucs_power.Power] + power



# Copyright 2013 Hewlett-Packard Development Company, L.P.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import time

from oslo_config import cfg

from ironic.common import exception
from ironic.common import hash_ring
from ironic.tests.unit.db import base as db_base

CONF = cfg.CONF


class HashRingManagerTestCase(db_base.DbTestCase):

    def setUp(self):
        super(HashRingManagerTestCase, self).setUp()
        self.ring_manager = hash_ring.HashRingManager()

    def register_conductors(self):
        c1 = self.dbapi.register_conductor({
            'hostname': 'host1',
            'drivers': ['driver1', 'driver2'],
        })
        c2 = self.dbapi.register_conductor({
            'hostname': 'host2',
            'drivers': ['driver1'],
        })
        for c in (c1, c2):
            self.dbapi.register_conductor_hardware_interfaces(
                c.id, 'hardware-type', 'deploy', ['iscsi', 'direct'], 'iscsi')

    def test_hash_ring_manager_get_ring_success(self):
        self.register_conductors()
        ring = self.ring_manager['driver1']
        self.assertEqual(sorted(['host1', 'host2']), sorted(ring.nodes))

    def test_hash_ring_manager_hardware_type_success(self):
        self.register_conductors()
        ring = self.ring_manager['hardware-type']
        self.assertEqual(sorted(['host1', 'host2']), sorted(ring.nodes))

    def test_hash_ring_manager_driver_not_found(self):
        self.register_conductors()
        self.assertRaises(exception.DriverNotFound,
                          self.ring_manager.__getitem__,
                          'driver3')

    def test_hash_ring_manager_automatic_retry(self):
        self.assertRaises(exception.TemporaryFailure,
                          self.ring_manager.__getitem__,
                          'driver1')
        self.register_conductors()
        self.assertIsNotNone(self.ring_manager['driver1'])

    def test_hash_ring_manager_reset_interval(self):
        CONF.set_override('hash_ring_reset_interval', 30)
        # Just to simplify calculations
        CONF.set_override('hash_partition_exponent', 0)
        self.dbapi.register_conductor({
            'hostname': 'host1',
            'drivers': ['driver1', 'driver2'],
        })

        ring = self.ring_manager['driver1']
        self.assertEqual(1, len(ring))

        self.dbapi.register_conductor({
            'hostname': 'host2',
            'drivers': ['driver1'],
        })
        ring = self.ring_manager['driver1']
        # The new conductor is not known yet. Automatic retry does not kick in,
        # since there is an active conductor for the requested driver.
        self.assertEqual(1, len(ring))

        self.ring_manager.updated_at = time.time() - 31
        ring = self.ring_manager['driver1']
        self.assertEqual(2, len(ring))



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import mock
from oslo_utils import uuidutils

from ironic.common import context
from ironic.common import driver_factory
from ironic.common import release_mappings
from ironic.db import api as db_api
from ironic.tests.unit.db import base
from ironic.tests.unit.db import utils


class UpgradingTestCase(base.DbTestCase):

    def setUp(self):
        super(UpgradingTestCase, self).setUp()
        self.dbapi = db_api.get_instance()
        self.object_versions = release_mappings.get_object_versions()

    def test_check_versions_emptyDB(self):
        # nothing in the DB
        self.assertTrue(self.dbapi.check_versions())

    def test_check_versions(self):
        for v in self.object_versions['Node']:
            node = utils.create_test_node(uuid=uuidutils.generate_uuid(),
                                          version=v)
            node = self.dbapi.get_node_by_id(node.id)
            self.assertEqual(v, node.version)
        self.assertTrue(self.dbapi.check_versions())

    def test_check_versions_node_no_version(self):
        node = utils.create_test_node(version=None)
        node = self.dbapi.get_node_by_id(node.id)
        self.assertIsNone(node.version)
        self.assertFalse(self.dbapi.check_versions())

    def test_check_versions_node_old(self):
        node = utils.create_test_node(version='1.0')
        node = self.dbapi.get_node_by_id(node.id)
        self.assertEqual('1.0', node.version)
        self.assertFalse(self.dbapi.check_versions())

    def test_check_versions_conductor(self):
        for v in self.object_versions['Conductor']:
            conductor = utils.create_test_conductor(
                uuid=uuidutils.generate_uuid(), version=v)
            conductor = self.dbapi.get_conductor(conductor.hostname)
            self.assertEqual(v, conductor.version)
        self.assertTrue(self.dbapi.check_versions())

    def test_check_versions_conductor_no_version(self):
        # works in Queens only
        conductor = utils.create_test_conductor(version=None)
        conductor = self.dbapi.get_conductor(conductor.hostname)
        self.assertIsNone(conductor.version)
        self.assertTrue(self.dbapi.check_versions())

    def test_check_versions_conductor_old(self):
        conductor = utils.create_test_conductor(version='1.0')
        conductor = self.dbapi.get_conductor(conductor.hostname)
        self.assertEqual('1.0', conductor.version)
        self.assertFalse(self.dbapi.check_versions())


class BackfillVersionTestCase(base.DbTestCase):

    def setUp(self):
        super(BackfillVersionTestCase, self).setUp()
        self.context = context.get_admin_context()
        self.dbapi = db_api.get_instance()
        obj_mapping = release_mappings.RELEASE_MAPPING['pike']['objects']
        self.conductor_ver = obj_mapping['Conductor'][0]

    def test_empty_db(self):
        self.assertEqual((0, 0),
                         self.dbapi.backfill_version_column(self.context, 10))

    def test_version_exists(self):
        utils.create_test_conductor()
        self.assertEqual((0, 0),
                         self.dbapi.backfill_version_column(self.context, 10))

    def test_one_conductor(self):
        conductors = self._create_conductors(1)
        self.assertEqual((1, 1),
                         self.dbapi.backfill_version_column(self.context, 10))
        res = self.dbapi.get_conductor(conductors[0])
        self.assertEqual(self.conductor_ver, res.version)

    def test_max_count_zero(self):
        conductors = self._create_conductors(2)
        self.assertEqual((2, 2),
                         self.dbapi.backfill_version_column(self.context, 0))
        for hostname in conductors:
            conductor = self.dbapi.get_conductor(hostname)
            self.assertEqual(self.conductor_ver, conductor.version)

    def _create_conductors(self, num, version=None):
        conductors = []
        for i in range(0, num):
            conductor = utils.create_test_conductor(
                version=version,
                hostname='test_name_%d' % i,
                uuid=uuidutils.generate_uuid())
            conductors.append(conductor.hostname)
        for hostname in conductors:
            conductor = self.dbapi.get_conductor(hostname)
            self.assertEqual(version, conductor.version)
        return conductors

    def test_no_version_max_count_2_some_conductors(self):
        conductors = self._create_conductors(5)

        self.assertEqual((5, 2),
                         self.dbapi.backfill_version_column(self.context, 2))
        self.assertEqual((3, 3),
                         self.dbapi.backfill_version_column(self.context, 10))
        for hostname in conductors:
            conductor = self.dbapi.get_conductor(hostname)
            self.assertEqual(self.conductor_ver, conductor.version)

    def test_no_version_max_count_same(self):
        conductors = self._create_conductors(5)

        self.assertEqual((5, 5),
                         self.dbapi.backfill_version_column(self.context, 5))
        for hostname in conductors:
            conductor = self.dbapi.get_conductor(hostname)
            self.assertEqual(self.conductor_ver, conductor.version)


@mock.patch.object(driver_factory, 'calculate_migration_delta', autospec=True)
@mock.patch.object(driver_factory, 'classic_drivers_to_migrate', autospec=True)
class MigrateToHardwareTypesTestCase(base.DbTestCase):

    def setUp(self):
        super(MigrateToHardwareTypesTestCase, self).setUp()
        self.context = context.get_admin_context()
        self.dbapi = db_api.get_instance()
        self.node = utils.create_test_node(uuid=uuidutils.generate_uuid(),
                                           driver='classic_driver')

    def test_migrate(self, mock_drivers, mock_delta):
        mock_drivers.return_value = {'classic_driver': mock.sentinel.drv1,
                                     'another_driver': mock.sentinel.drv2}
        mock_delta.return_value = {'driver': 'new_driver',
                                   'inspect_interface': 'new_inspect'}
        result = self.dbapi.migrate_to_hardware_types(self.context, 0)
        self.assertEqual((1, 1), result)
        node = self.dbapi.get_node_by_id(self.node.id)
        self.assertEqual('new_driver', node.driver)
        self.assertEqual('new_inspect', node.inspect_interface)

    def test_migrate_unsupported(self, mock_drivers, mock_delta):
        mock_drivers.return_value = {'classic_driver': mock.sentinel.drv1,
                                     'another_driver': mock.sentinel.drv2}
        mock_delta.return_value = None
        result = self.dbapi.migrate_to_hardware_types(self.context, 0)
        self.assertEqual((1, 1), result)
        node = self.dbapi.get_node_by_id(self.node.id)
        self.assertEqual('classic_driver', node.driver)


class UpdateToLatestVersionsTestCase(base.DbTestCase):

    def setUp(self):
        super(UpdateToLatestVersionsTestCase, self).setUp()
        self.context = context.get_admin_context()
        self.dbapi = db_api.get_instance()

        obj_versions = release_mappings.get_object_versions(
            objects=['Node', 'Chassis'])
        master_objs = release_mappings.RELEASE_MAPPING['master']['objects']
        self.node_ver = master_objs['Node'][0]
        self.chassis_ver = master_objs['Chassis'][0]
        self.node_old_ver = self._get_old_object_version(
            self.node_ver, obj_versions['Node'])
        self.chassis_old_ver = self._get_old_object_version(
            self.chassis_ver, obj_versions['Chassis'])
        self.node_version_same = self.node_old_ver == self.node_ver
        self.chassis_version_same = self.chassis_old_ver == self.chassis_ver
        # number of objects with different versions
        self.num_diff_objs = 2
        if self.node_version_same:
            self.num_diff_objs -= 1
        if self.chassis_version_same:
            self.num_diff_objs -= 1

    def _get_old_object_version(self, latest_version, versions):
        """Return a version that is older (not same) as latest version.

        If there aren't any older versions, return the latest version.
        """
        for v in versions:
            if v != latest_version:
                return v
        return latest_version

    def test_empty_db(self):
        self.assertEqual(
            (0, 0), self.dbapi.update_to_latest_versions(self.context, 10))

    def test_version_exists(self):
        # Node will be in latest version
        utils.create_test_node()
        self.assertEqual(
            (0, 0), self.dbapi.update_to_latest_versions(self.context, 10))

    def test_one_node(self):
        node = utils.create_test_node(version=self.node_old_ver)
        expected = (0, 0) if self.node_version_same else (1, 1)
        self.assertEqual(
            expected, self.dbapi.update_to_latest_versions(self.context, 10))
        res = self.dbapi.get_node_by_uuid(node.uuid)
        self.assertEqual(self.node_ver, res.version)

    def test_max_count_zero(self):
        orig_node = utils.create_test_node(version=self.node_old_ver)
        orig_chassis = utils.create_test_chassis(version=self.chassis_old_ver)
        self.assertEqual((self.num_diff_objs, self.num_diff_objs),
                         self.dbapi.update_to_latest_versions(self.context, 0))
        node = self.dbapi.get_node_by_uuid(orig_node.uuid)
        self.assertEqual(self.node_ver, node.version)
        chassis = self.dbapi.get_chassis_by_uuid(orig_chassis.uuid)
        self.assertEqual(self.chassis_ver, chassis.version)

    def test_old_version_max_count_1(self):
        orig_node = utils.create_test_node(version=self.node_old_ver)
        orig_chassis = utils.create_test_chassis(version=self.chassis_old_ver)
        num_modified = 1 if self.num_diff_objs else 0
        self.assertEqual((self.num_diff_objs, num_modified),
                         self.dbapi.update_to_latest_versions(self.context, 1))
        node = self.dbapi.get_node_by_uuid(orig_node.uuid)
        chassis = self.dbapi.get_chassis_by_uuid(orig_chassis.uuid)
        self.assertTrue(node.version == self.node_old_ver or
                        chassis.version == self.chassis_old_ver)
        self.assertTrue(node.version == self.node_ver or
                        chassis.version == self.chassis_ver)

    def _create_nodes(self, num_nodes):
        version = self.node_old_ver
        nodes = []
        for i in range(0, num_nodes):
            node = utils.create_test_node(version=version,
                                          uuid=uuidutils.generate_uuid())
            nodes.append(node.uuid)
        for uuid in nodes:
            node = self.dbapi.get_node_by_uuid(uuid)
            self.assertEqual(version, node.version)
        return nodes

    def test_old_version_max_count_2_some_nodes(self):
        if self.node_version_same:
            # can't test if we don't have diff versions of the node
            return

        nodes = self._create_nodes(5)
        self.assertEqual(
            (5, 2), self.dbapi.update_to_latest_versions(self.context, 2))
        self.assertEqual(
            (3, 3), self.dbapi.update_to_latest_versions(self.context, 10))
        for uuid in nodes:
            node = self.dbapi.get_node_by_uuid(uuid)
            self.assertEqual(self.node_ver, node.version)

    def test_old_version_max_count_same_nodes(self):
        if self.node_version_same:
            # can't test if we don't have diff versions of the node
            return

        nodes = self._create_nodes(5)
        self.assertEqual(
            (5, 5), self.dbapi.update_to_latest_versions(self.context, 5))
        for uuid in nodes:
            node = self.dbapi.get_node_by_uuid(uuid)
            self.assertEqual(self.node_ver, node.version)
